// usage: node schemaBackup.js config.json

const { execSync } = require('node:child_process');
const {NodeSSH} = require('node-ssh')
var bunyan = require('bunyan');
var Promise = require("bluebird");
const spawnSync = require('child_process').spawnSync;

const log = bunyan.createLogger({
    name: 'schemaBackup',
    streams: [
      {
        level: 'debug',
        stream: process.stdout 
      }
    ]
  });

//input
let param = process.argv
log.info('process.argv：\n', param)  
//input params
const configFile = (param.length > 2)?param[2]:'config.json'
const configure = require(`${configFile}`);
//input end

log.info('configure',configure)
const k8sNs =  configure.k8sNs?configure.k8sNs:'all'
const exec = true
const type = configure.type?configure.type:'all'
const excludeMongoLarge = configure.excludeMongoLarge
const backPathMongo = configure.backPathMongo?configure.backPathMongo:'/data_disk/nfsmount/mongoBackup' 
const backPathLocal = `${backPathMongo}/${k8sNs}`
//pg
const dbName = configure.dbName?configure.dbName:false
const pgDir = configure.pgDir?configure.pgDir:'/usr/pgsql-14/bin'
const backPathLocalPg = configure.backPathLocalPg?configure.backPathLocalPg:'/data_disk/pgBackup'
//hana
const backupFrequency = configure.backupFrequency?configure.backupFrequency:'hour'

const HANA_SYSTEM_ACCOUNT = 'SYSTEM'
const HANA_SYSTEM_PASSWORD = 'Wang_2017'
const HANA_PORT = 30115
const HANA_BACKUPPATH = 'nfsmount'
const HANA_SCHEMA_BACKUP_PATH = 'SCHEMA_BACKUP'
const HANA_DB_BACKUP_PATH = 'DB_BACKUP'

let hanaCnt = 0
let mongoCnt = 0
let hanaTotal = 0
let mongoTotal = 0
let pgTotal = 0
let pgCnt = 0

let hour,day,month,dateStr = null

const { Client } = require('pg')
//from CMDB
let cfg = {
    user: 'postgres',
    host: 'hanatest1.epmtest.cn',
    database: 'cmdb',
    password: 'Wang_2023',
    port: 5432,
  }
const client = new Client(cfg)

const hana = require('hdb')



function arraySql(dbList) {
  let ary = dbList.split(',')
  let r = '('
  for (let t of ary) {
     r = r + `'${t}',`
  }
  log.debug(r)
  r = r.slice(0,-1)
  r = r + ')'
  return r
}


async function pgBackup() {  
  let sqlName = `select * from service_info_pg where sys_envid != 419 and open = true`  
  if (dbName) {
    sqlName = sqlName + ` and dbname in ${arraySql(dbName)}`
  }
  if (backupFrequency) {
    sqlName = sqlName + ` and backup_frequency = '${backupFrequency}'`
  }
  log.info(sqlName);                     
  let r = await client.query(sqlName)
  let list = r.rows
  log.debug(list)
  pgTotal = list.length

  list.forEach(async (m) => { //iterate over list
    let dbName = ''
    try {
      log.debug(m)
      let ip = m.pg_ip
      let port = m.pg_port
      dbName = m.dbname

      //get dump name from time
      let dumpName = `${dbName}_${month}_${date}_${hour}.c`
      log.info(`pg info ${ip}:${port}, ${dbName},dumpName ${dumpName}`)

      // /usr/pgsql-14/bin/pg_dump -h hanatest1.epmtest.cn -p 5432 -U postgres -n public yatal > /data_disk/pgBackup/yatal.dump
      let cmd = `${pgDir}/pg_dump -h ${ip} -p ${port} -U postgres -n public ${dbName} -Fc -f ${backPathLocalPg}/${dumpName}`
      let cmdOps = []
      let rsp = spawnSync(cmd, cmdOps,{shell: true,encoding: 'utf-8', stdio: 'pipe'});

      log.debug(cmd)
      // will prompt for password
      // let cmdOps = ['-h', ip,'-p',port,'-U postgres','-n public',dbName,'-f',`${backPathLocalPg}/${dbName}.dump`]
      // log.info(cmdOps)
      // let rsp = spawnSync('/usr/pgsql-14/bin/pg_dump', cmdOps,{shell: true,encoding: 'utf-8', stdio: 'pipe'}); 
      log.debug(rsp)
      if (rsp.stderr) {
        throw rsp.stderr
      }
      log.info('dump success')
               
      log.info(`backup done for ${dbName}`)
      
    } catch (e) {
      log.error(`error processing ${dbName}`,e)
    }

    if ( ++pgCnt >= pgTotal ) {
      log.info('all finished')
      process.exit(0)
    }

  }) //forEach

  log.info('pg backup done async')    

}


async function hanaBackup() {
  let sqlHanaName = 'select * from hana_info where env_ywid != 419' 
  //let sqlHanaName = 'select * from hana_info where id = 5'  //debug poc3
  log.info(sqlHanaName);                     
  let r = await client.query(sqlHanaName)
  let hanaList = r.rows
  log.debug(hanaList)
  hanaTotal = hanaList.length

  hanaList.forEach(async (m) => { //iterate over hana,paral
    let hanaName = ''
    let dbBUPath = '' //for db
    //let dataPath = '' //for ecs
    let ip = ''
    let sql = ''

    try {
      //let hanaId = m.id
      let dockerDeploy = m.docker_deploy
      hanaName = m.name
      if (backupFrequency != 'hour' ) {
        log.info('data backup DB start, for ecs hana:' + hanaName)
        dbBUPath = m.data_backup_path
                
        let defaultDbname = dockerDeploy?'hxe':'epm'
        let dbName = m.dbname?m.dbname:defaultDbname        
        
        let configSys = {
          "host": m.ip,
          "port": HANA_PORT,
          "user": HANA_SYSTEM_ACCOUNT,
          "schema": HANA_SYSTEM_ACCOUNT,
          "password": HANA_SYSTEM_PASSWORD,
        };
        log.debug(JSON.stringify(configSys));
        let hdbClientSys = hana.createClient(configSys);
        Promise.promisifyAll(hdbClientSys)
        await hdbClientSys.connectAsync()
        log.debug('hdbClientSys connectAsync done');
        sql = `BACKUP DATA using file ('${dbBUPath}/${dbName}_${dateStr}')`
        log.info(sql)
        await hdbClientSys.execAsync(sql) 
        log.info(`DB backup done for hana ${hanaName}`)
      }

      log.info('backup schema')      
      let sqlSchema = `select * from schema_info where name = '${hanaName}' and open = true` 
      if (backupFrequency) {
        sqlSchema = sqlSchema + ` and (backup_frequency = '${backupFrequency}' or env_ywid = 2)`
      }      
      log.info(sqlSchema);                     
      let r = await client.query(sqlSchema)
      log.debug(r)  
      let schemaList = r.rows
      log.info('schemaList',schemaList)

      for (let r1 of schemaList) { //iterate over schema          
        log.debug(r1)
        let hanaName = ''
        let port = ''
        let schemaName = ''

        try {
            schemaName = r1.hana_schema
            let backPath = r1.backup_path
            if (backupFrequency == 'hour') {
              backPath = backPath + '_HOUR'
            }
            // dataPath = r1.backup_ecs_path
            // if (backupFrequency == 'hour') {
            //   dataPath = dataPath + '/HOURLY'
            // }
            ip = r1.ip
            hanaName = r1.name
            port = r1.port
            let password = r1.schema_password
            log.info(`processing hana ${hanaName}  schemaName ${schemaName}`)
            let config = {
                "host": ip,
                "port": port,
                "user": schemaName,
                "schema": schemaName,
                "password": password,
            };
            log.debug(JSON.stringify(config));
            const hdbClient = hana.createClient(config);
            Promise.promisifyAll(hdbClient)
            await hdbClient.connectAsync()
            log.debug('hdbClient connectAsync done');

            // hdbsql -u $schema -p EPM_user_2017 -n $remoteHana "EXPORT $schema.\"*\" AS BINARY INTO '$backupDir/$dateStr/$schema' WITH REPLACE"
            sql = `export ${schemaName}."*" as binary into '${backPath}/${schemaName}_${dateStr}' with replace`
            log.info(sql)
            if (exec) {
              await hdbClient.execAsync(sql) 
            }
            log.info(`schema backup done for hana ${hanaName}, schema ${schemaName}`)

            // ssh to leondev, cp nfs to nas
            log.info(' ssh to leondev, cp nfs to nas')
            const ssh = new NodeSSH()
            await ssh.connect({
              host: 'cicdcache.epmtest.cn',
              port: 40009,
              username: 'epm',
              privateKeyPath: '/home/epm/.ssh/id_rsa'
            })  
            let cmd = `cd /mnt/nfs/prod2/SCHEMA_BACKUP_HOUR;sudo tar zcf ${schemaName}_${dateStr}.tar ${schemaName}_${dateStr};sudo cp ${schemaName}_${dateStr}.tar /mnt/nas/DB_BACKUP/hanaProd2/SCHEMA_BACKUP`
            log.debug(cmd)
            r = await ssh.execCommand(cmd)
            if (r.code > 0) {
              log.error('error',r.stderr)
            } else {
                log.info('good',r.stdout) //result
            }     

        } catch (e) {
            log.error(`backup error for hana ${hanaName}, schema ${schemaName}`,e)
        }  
      

      }//for schema list
      //log.info(`backup DB done for hana ${hanaName}`)

      //clean backup: moved to local crontab job
      // const ssh = new NodeSSH()
      // await ssh.connect({
      //   host: ip,
      //   port: 22,
      //   username: 'epm',
      //   privateKeyPath: '/home/epm/.ssh/id_rsa'
      // })  
      // let reserveDay = (k8sNs == 'test')?7:14
      // let cleanCmd = `sudo find ${dataPath} -type d -mtime +${reserveDay} -exec rm -r {} \\\;`
      // log.debug(cleanCmd)
      // r = await ssh.execCommand(cleanCmd)
      // log.debug(r)      
      // // hourly backup clean
      // cleanCmd = `sudo find ${dataPath} -type d -mtime +2 -exec rm -r {} \\\;`
      // log.debug(cleanCmd)
      // r = await ssh.execCommand(cleanCmd)
      // log.debug(r)      
    } catch (e) {
      log.error(`${hanaName} bad`,e)
    }

    if ( ++hanaCnt >= hanaTotal ) {
      log.info(`hana backup all finished`)
      process.exit(0)
    }
  }) //for hanaList.forEach

  log.info('hanaBackup done async')  

}

async function mongoBackup1(m) {
  try {

      log.debug(m)
      let mongoName = m.name
      let sqlSchema = `select * from schema_info where mongo_name = '${mongoName}' and open = true`       
      log.info(sqlSchema);                     
      r = await client.query(sqlSchema)
      let schemaList = r.rows
      log.debug(schemaList)
      let ip = m.ip
      let port = m.port

      for (let r1 of schemaList) { //iterate over schema          
        log.debug(r1)
        
        let schemaName = ''   
    
        try {
            schemaName = r1.mongo_schema
            //backPath = backPathLocal + '/' + schemaName
            let password = r1.mongo_password
            log.info(`processing mongoName ${mongoName}  schemaName ${schemaName}`)
             
            //mongodump -h 124.220.100.10 --port 27018 -u YATALCOPY -p EPM_user_2017 -d YATALCOPY -o /home/epm/mongoBackup
            if (exec) {
              let cmd = `sudo mongodump -h ${ip} --port ${port} -u ${schemaName} -p ${password} -d ${schemaName} -o ${backPathLocal}/${dateStr}`
              if (excludeMongoLarge) {
                cmd = cmd + ' --excludeCollection=logCollection --excludeCollection=etlLogCollection --excludeCollection=jounalCollection'
              }
              log.info(cmd)
              let r = execSync(cmd)
              log.debug(r)
              log.info('mongodump success')
            }            
            log.info(`backup done for mongoName ${mongoName}  schemaName ${schemaName}`)

        } catch (e) {
            log.error(`error processing mongoName ${mongoName}  schemaName ${schemaName}`,e)
        }  

      }//for schema list
      log.info(`backup all mongo dbs for ${mongoName}`)

  } catch(e) {
      log.error()
  }

}


async function mongoBackup() {
  
  let sqlName = `select * from mongo_info` 
  if (k8sNs != 'all') {
    sqlName = sqlName + ` where k8s_namespace='${k8sNs}'`
  }
  log.info(sqlName);                     
  let r = await client.query(sqlName)
  let list = r.rows
  log.debug(list)
  const queuedTasks =[];
  list.forEach(e => {
      const subTask = mongoBackup1(e)
      queuedTasks.push(subTask);                    
  });               
  log.debug('queuedTasks.length',queuedTasks.length)
  await Promise.all(queuedTasks); 
  // log.info('syncTecnetSchema all done')  
  let msg = `mongoBackup all finished`
  log.info(msg)

  // ssh to leondev, cp nfs to nas
  log.info(' ssh to leondev, cp nfs to nas')
  const ssh = new NodeSSH()
  await ssh.connect({
    host: 'cicdcache.epmtest.cn',
    port: 40009,
    username: 'epm',
    privateKeyPath: '/home/epm/.ssh/id_rsa'
  })  
  let cmd = `cd /mnt/nfs/mongoBackup/epmprod;sudo cp -r ${dateStr} /mnt/nas/DB_BACKUP/mongoProd`
  log.debug(cmd)
  r = await ssh.execCommand(cmd)
  if (r.code > 0) {
    log.error('error',r.stderr)
  } else {
      log.info('to nas good',r.stdout) //result
  }     


}

//main
;( async () => {
  //get utc time
  var dateObj = new Date();
  hour = dateObj.getHours()
  date = dateObj.getDate()
  month = dateObj.getMonth() + 1
  dateStr = `${month}_${date}`
  if ( backupFrequency == 'hour') {
    dateStr = dateStr + `_${hour}`
  }

   log.info(`month ${month} day ${date} hour ${hour}`)
    
    log.info('read from CMDB') 
    await client.connect()

    if (type == 'all' || type == 'pg') {
      await pgBackup()
    }

    if (type == 'all' || type == 'hana') {
      await hanaBackup()     
    }

    if (type == 'all' || type == 'mongo') {
      await mongoBackup()
    }

  })().catch(err => {
    log.error('main exit ',err.stack)
    process.exit(1)
})