var bunyan = require('bunyan');
var pgLog = bunyan.createLogger({
    name: 'pgClient',
    streams: [       
          {
            level: 'debug',
            stream: process.stdout
          },
    ]
  });

const dtMap = {
    "nvarchar":"varchar",
    "nclob":"text",
    "longdate":"timestamp",
    "blob":"bytea",
    "tinyint":"smallint"
}
//convert column,table name in "" to lower case
function toLowerSql(sql) {
    let patt = /"\S+"/g;
    if (patt.test(sql)) {
        let ary=sql.match(patt);
        ary.forEach(e => {
            sql = sql.replace(e,e.toLowerCase())        
        });
    }
    return sql
}

function convertHanaAlterColumnToPg(sql) {
    //TODO: comment supported , alter supported
    let newSql = sql;
    let defaultSql = sql;
    let firstParenthesisIdx = defaultSql.indexOf('(');
    if (firstParenthesisIdx > 0) {
        //alter table tablename add|drop[  ](
        let left = defaultSql.substring(0, firstParenthesisIdx + 1);

        // columnDefs )
        let right = defaultSql.substring(firstParenthesisIdx + 1);
        let i = right.lastIndexOf(')')
        if (i > 0) {
            right = right.substring(0, i)
        }
        let actionRegex = /\s(add|drop)\s*\(/i;
        let actionMatch = actionRegex.exec(left);
        if (actionMatch !== null) {
            let action = actionMatch[1];//action with pre space
            action = action.trim();
            let leftWithoutAction = left.substring(0, actionMatch.index);//remove add|drop
            let commaRegex = /,(?!\s*\d+\s*\))/gi;///(?<!,[^(]+\([^)]+),/;
            let rightPart = right.replace(commaRegex,`,${action} `);
            newSql = leftWithoutAction + " " + action + " " + rightPart;
        }
    }
    return newSql;
}

//convert hana to pg sql (DDL/DML)
function processSql(sql) {
    pgLog.debug(`original sql:${sql}`)
   
    //common process
    //convert keyword in "" to lower case
    sql = toLowerSql(sql)
    sql = sql.replace(/UNLOAD\s+PRIORITY\s+\d+\s+AUTO\s+MERGE/gi,'')     
    sql = sql.replace(/CPBTREE/gi,'') 
    sql = sql.replace(/current_utctimestamp/gi,'current_timestamp')     
    sql = sql.replace(/IFNULL/gi,'COALESCE')     
    //to_varchar(a) as b   => cast(a as varchar) as b            
    let patt = /TO_VARCHAR\s*\(\s*\S+\s*\)/gi;
    if (patt.test(sql)) {
        let ary=sql.match(patt);
        //console.log(ary)
        ary.forEach(e => {
            let patt1 = /TO_VARCHAR\s*\(\s*(?<colName>\S+)\s*\)/gi;
            //console.log(e)
            let colName = patt1.exec(e).groups.colName;
            //console.log(colName)
            let rep = `cast(${colName} as varchar)`
            sql = sql.replace(e,rep)    
        });
    }   
    
    //DDL
    patt = /(CREATE|ALTER).*(TABLE|VIEW)\s+"*\w+"*/i;
    if (patt.test(sql)) {
        //DDL may contain default values
        //sql = lowerSql
        //simple replace
        sql = sql.replace(/\s+column\s+/gi,' ')
        //sql = sql.replace(/"column"/gi,'')
        sql = sql.replace(/\s+row\s+/gi,' ')
        //remove MEMORY THRESHOLD
        sql = sql.replace(/\s+MEMORY THRESHOLD\s+\d+/gi,' ')
        //remove INVERTED
        sql = sql.replace(/\s+INVERTED\s+VALUE\s+/gi,' ')
        //remove WITH READ ONLY
        sql = sql.replace(/\s+WITH\s+READ\s+ONLY/gi,' ')
        //remove CS_LONGDATE
        sql = sql.replace(/\s+cs_\w+/gi,' ')

        //pg key word add ""
        // kwList.forEach(k => {
        //     let pt =new RegExp(`\\s+${k}\\s+`,'gi')
        //     sql = sql.replace(pt,` "${k}" `)
        // })

        //ddl without data type            
        //process like with data            
        if (sql.toLowerCase().includes('with data')) {
            sql = sql.replace(/\s+like\s+/i,' as table ')
        }
        //process like
        // create table T_U1_C21_c LIKE T_U1_C21
        //pg: create table T_U1_C21_c (LIKE T_U1_C21)
        let pt1 = /(create.*table.*)\s+(like)\s+("?\w+"?)(.*)/i;
        if (pt1.test(sql)) {
            sql = sql.toLowerCase();
            let matches = pt1.exec(sql);
            sql = matches[1]+" ("+matches[2]+" "+matches[3]+" INCLUDING INDEXES ) "+matches[4];
        } else {
            //data type tansform for ddl
            for (let hanaType in dtMap) {
                let pgType = dtMap[hanaType]
                let p =new RegExp(hanaType, 'gi')
                sql = sql.replace(p,pgType)
            }          
            
            //ALTER TABLE distributors ADD COLUMN address varchar(30);
            //alter table add/drop (column),alter table alter()
            //alter table tableName add constraint [***_pkey] unique("c1")
            if(/alter\s+table.*/i.test(sql)){
                //let constraintRegex = /alter\s+table\s+.*(drop|add)\s+constraint?/i;
                let columnRegx = /alter\s+table\s+.*(drop|add)\s*\(/i;
                if (columnRegx.test(sql)) {
                    sql = convertHanaAlterColumnToPg(sql);
                } 
            }

            //ALTER TABLE distributors ADD COLUMN address varchar(30);
            // if (sql.toLowerCase().includes('alter table')) {
            //     //remove outest ()
            //     let t = sql.replace(/\(/,' ')
            //     let i = t.lastIndexOf(')')
            //     if (i > 0) { 
            //         sql = t.slice(0,i)
            //     }
            // }     
        }
        pgLog.debug('exec:' + sql)
    } else {
        //DML
         //process TREE expression
        let pattTree = /\s+from\s+\w+\s+\(\s*expression\s*=>\s*'\w+\(\s*\S+\s*\)'\)/gi;
        if (pattTree.test(sql)) {
            pgLog.debug('process TREE expression')
    
            let ary=sql.match(pattTree);
            pgLog.debug(ary)
            ary.forEach(e => {
            // console.log(e)
            let patt = /\s+from\s+(?<viewName>\w+)\s+\(\s*expression\s*=>\s*'(?<funcName>\w+)\(\s*(?<paramName>\S+)\s*\)'\)/gi;
                let regGrp = patt.exec(e).groups
                //console.log(regGrp)
                let v = regGrp.viewName
                let f = regGrp.funcName
                let p = regGrp.paramName
                //RESULT_NODE ROM LEAVEs('H_U4_AUDIT',100119363)
                let repS = ` FROM ${f}('${v}',${p})`
                sql = sql.replace(e,repS)
            });
    
            pgLog.debug('tree expression transformed:' + sql)        
        }    
        
        //process update from
        //HANA: UPDATE T_TEST0717_JCON_0822_1 AS T SET T.R_GUANXI0825=R.VALUE FROM T_TEST0717_JCON_0822_1 AS T, RELATIONSHIP_VAL AS R WHERE T.D_JEN=R.DIM1_ID AND T.D_JINT=R.DIM2_ID AND R.CODE=GUANXI0825 and R.UNIVERSE=TEST0717 and R.UCUBE=JCON_0822_1
        //PG: UPDATE T_TEST0717_JCON_0822_1 AS T SET R_GUANXI0825=R.VALUE FROM RELATIONSHIP_VAL AS R 
        // WHERE T.D_JEN=R.DIM1_ID AND T.D_JINT=R.DIM2_ID AND R.CODE='GUANXI0825' and R.UNIVERSE='TEST0717' and R.UCUBE='JCON_0822_1'
        let patt = /update\s+(?<tblName>\w+)\s+as\s+(?<tblAlise>\w+)\s+set\s+(?<setClause>.+)\s+from\s+(?<fromClause>.+)\s+where\s+(?<whereClause>.+)/i;

        if (patt.test(sql)) {
            pgLog.debug('process update from')
            let grp = patt.exec(sql).groups
            pgLog.debug(grp)
            let setC = grp.setClause.replace(`${grp.tblAlise}\.`,'')
            let fromRe = new RegExp(`${grp.tblName} as ${grp.tblAlise}`,'i');
            let fromC = grp.fromClause.replace(fromRe,'')
            fromC = fromC.replace(',','')
            sql = `update ${grp.tblName} as ${grp.tblAlise} set ${setC} from ${fromC} where ${grp.whereClause}`
            pgLog.debug('pgsql: ' + sql)
        }

        //data type tansform
        for (let hanaType in dtMap) {
            let pgType = dtMap[hanaType]
            let p =new RegExp(hanaType, 'gi')
            sql = sql.replace(p,pgType)
        }
        //handle select from sequences
        patt = /from\s+(sequences|m_sequences)/i;
        if (patt.test(sql)) {
            // sequence object must be in lower case
            sql = sql.toLowerCase();
            sql = sql.replace(/from\s+(sequences|m_sequences)/,'from information_schema.sequences')
        }         
        //handle set schema
        sql = sql.replace(/set\s+schema/i,'set search_path to')            

        //sql = sql.replace(/STRING_AGG/gi,'concat')            
        
        sql = sql.replace(/\s+from\s+DUMMY\s*/gi,'')

        patt = /\w+\.nextval/gi;
        if (patt.test(sql)) {
            let ary=sql.match(patt);
            ary.forEach(e => {
                let patt1 = /(?<objName>\w+)\.nextval/i;
                let objName = patt1.exec(e).groups.objName;
                let rep = `nextval('${objName}')`
                sql = sql.replace(e,rep) 
            });
        }

    }
    return sql
}


module.exports = {processSql};
//export default { processSql };